/*
 * This file is automagically created from term.c -- DO NOT EDIT
 */

#define T_FLG 0
#define T_NUM 1
#define T_STR 2

struct term
{
  char *tcname;
  int type;
};

union tcu
{
  int flg;
  int num;
  char *str;
};

#define d_LI  d_tcs[0].num
#define D_LI (D_tcs[0].num)
#define d_CO  d_tcs[1].num
#define D_CO (D_tcs[1].num)
#define d_HC  d_tcs[2].flg
#define D_HC (D_tcs[2].flg)
#define d_OS  d_tcs[3].flg
#define D_OS (D_tcs[3].flg)
#define d_NS  d_tcs[4].flg
#define D_NS (D_tcs[4].flg)
#define d_CM  d_tcs[5].str
#define D_CM (D_tcs[5].str)
#define d_HO  d_tcs[6].str
#define D_HO (D_tcs[6].str)
#define d_CR  d_tcs[7].str
#define D_CR (D_tcs[7].str)
#define d_UP  d_tcs[8].str
#define D_UP (D_tcs[8].str)
#define d_CUP  d_tcs[9].str
#define D_CUP (D_tcs[9].str)
#define d_DO  d_tcs[10].str
#define D_DO (D_tcs[10].str)
#define d_CDO  d_tcs[11].str
#define D_CDO (D_tcs[11].str)
#define d_BS  d_tcs[12].flg
#define D_BS (D_tcs[12].flg)
#define d_BC  d_tcs[13].str
#define D_BC (D_tcs[13].str)
#define d_LE  d_tcs[14].str
#define D_LE (D_tcs[14].str)
#define d_CLE  d_tcs[15].str
#define D_CLE (D_tcs[15].str)
#define d_ND  d_tcs[16].str
#define D_ND (D_tcs[16].str)
#define d_CRI  d_tcs[17].str
#define D_CRI (D_tcs[17].str)
#define d_CS  d_tcs[18].str
#define D_CS (D_tcs[18].str)
#define d_NL  d_tcs[19].str
#define D_NL (D_tcs[19].str)
#define d_SF  d_tcs[20].str
#define D_SF (D_tcs[20].str)
#define d_SR  d_tcs[21].str
#define D_SR (D_tcs[21].str)
#define d_AL  d_tcs[22].str
#define D_AL (D_tcs[22].str)
#define d_CAL  d_tcs[23].str
#define D_CAL (D_tcs[23].str)
#define d_DL  d_tcs[24].str
#define D_DL (D_tcs[24].str)
#define d_CDL  d_tcs[25].str
#define D_CDL (D_tcs[25].str)
#define d_IN  d_tcs[26].flg
#define D_IN (D_tcs[26].flg)
#define d_IM  d_tcs[27].str
#define D_IM (D_tcs[27].str)
#define d_EI  d_tcs[28].str
#define D_EI (D_tcs[28].str)
#define d_IC  d_tcs[29].str
#define D_IC (D_tcs[29].str)
#define d_CIC  d_tcs[30].str
#define D_CIC (D_tcs[30].str)
#define d_DC  d_tcs[31].str
#define D_DC (D_tcs[31].str)
#define d_CDC  d_tcs[32].str
#define D_CDC (D_tcs[32].str)
#define d_UT  d_tcs[33].flg
#define D_UT (D_tcs[33].flg)
#define d_CL  d_tcs[34].str
#define D_CL (D_tcs[34].str)
#define d_CD  d_tcs[35].str
#define D_CD (D_tcs[35].str)
#define d_CCD  d_tcs[36].str
#define D_CCD (D_tcs[36].str)
#define d_CE  d_tcs[37].str
#define D_CE (D_tcs[37].str)
#define d_CB  d_tcs[38].str
#define D_CB (D_tcs[38].str)
#define d_IS  d_tcs[39].str
#define D_IS (D_tcs[39].str)
#define d_TI  d_tcs[40].str
#define D_TI (D_tcs[40].str)
#define d_TE  d_tcs[41].str
#define D_TE (D_tcs[41].str)
#define d_BL  d_tcs[42].str
#define D_BL (D_tcs[42].str)
#define d_VB  d_tcs[43].str
#define D_VB (D_tcs[43].str)
#define d_CWS  d_tcs[44].str
#define D_CWS (D_tcs[44].str)
#define d_CZ0  d_tcs[45].str
#define D_CZ0 (D_tcs[45].str)
#define d_CZ1  d_tcs[46].str
#define D_CZ1 (D_tcs[46].str)
#define T_ATTR 47
#define d_MH  d_tcs[47].str
#define D_MH (D_tcs[47].str)
#define d_US  d_tcs[48].str
#define D_US (D_tcs[48].str)
#define d_MD  d_tcs[49].str
#define D_MD (D_tcs[49].str)
#define d_MR  d_tcs[50].str
#define D_MR (D_tcs[50].str)
#define d_SO  d_tcs[51].str
#define D_SO (D_tcs[51].str)
#define d_MB  d_tcs[52].str
#define D_MB (D_tcs[52].str)
#define d_UE  d_tcs[53].str
#define D_UE (D_tcs[53].str)
#define d_SE  d_tcs[54].str
#define D_SE (D_tcs[54].str)
#define d_ME  d_tcs[55].str
#define D_ME (D_tcs[55].str)
#define d_MS  d_tcs[56].flg
#define D_MS (D_tcs[56].flg)
#define d_SG  d_tcs[57].num
#define D_SG (D_tcs[57].num)
#define d_UG  d_tcs[58].num
#define D_UG (D_tcs[58].num)
#define d_SA  d_tcs[59].str
#define D_SA (D_tcs[59].str)
#define d_CAF  d_tcs[60].str
#define D_CAF (D_tcs[60].str)
#define d_CAB  d_tcs[61].str
#define D_CAB (D_tcs[61].str)
#define d_CSF  d_tcs[62].str
#define D_CSF (D_tcs[62].str)
#define d_CSB  d_tcs[63].str
#define D_CSB (D_tcs[63].str)
#define d_OP  d_tcs[64].str
#define D_OP (D_tcs[64].str)
#define d_CCO  d_tcs[65].num
#define D_CCO (D_tcs[65].num)
#define d_BE  d_tcs[66].flg
#define D_BE (D_tcs[66].flg)
#define d_CAX  d_tcs[67].flg
#define D_CAX (D_tcs[67].flg)
#define d_CC8  d_tcs[68].flg
#define D_CC8 (D_tcs[68].flg)
#define d_KS  d_tcs[69].str
#define D_KS (D_tcs[69].str)
#define d_KE  d_tcs[70].str
#define D_KE (D_tcs[70].str)
#define d_CCS  d_tcs[71].str
#define D_CCS (D_tcs[71].str)
#define d_CCE  d_tcs[72].str
#define D_CCE (D_tcs[72].str)
#define d_PO  d_tcs[73].str
#define D_PO (D_tcs[73].str)
#define d_PF  d_tcs[74].str
#define D_PF (D_tcs[74].str)
#define d_HS  d_tcs[75].flg
#define D_HS (D_tcs[75].flg)
#define d_WS  d_tcs[76].num
#define D_WS (D_tcs[76].num)
#define d_TS  d_tcs[77].str
#define D_TS (D_tcs[77].str)
#define d_FS  d_tcs[78].str
#define D_FS (D_tcs[78].str)
#define d_DS  d_tcs[79].str
#define D_DS (D_tcs[79].str)
#define d_VI  d_tcs[80].str
#define D_VI (D_tcs[80].str)
#define d_VS  d_tcs[81].str
#define D_VS (D_tcs[81].str)
#define d_VE  d_tcs[82].str
#define D_VE (D_tcs[82].str)
#define d_AM  d_tcs[83].flg
#define D_AM (D_tcs[83].flg)
#define d_XV  d_tcs[84].flg
#define D_XV (D_tcs[84].flg)
#define d_XN  d_tcs[85].flg
#define D_XN (D_tcs[85].flg)
#define d_COP  d_tcs[86].flg
#define D_COP (D_tcs[86].flg)
#define d_CLP  d_tcs[87].flg
#define D_CLP (D_tcs[87].flg)
#define d_CNF  d_tcs[88].flg
#define D_CNF (D_tcs[88].flg)
#define d_NX  d_tcs[89].flg
#define D_NX (D_tcs[89].flg)
#define d_CAN  d_tcs[90].flg
#define D_CAN (D_tcs[90].flg)
#define d_COL  d_tcs[91].num
#define D_COL (D_tcs[91].num)
#define d_CKJ  d_tcs[92].str
#define D_CKJ (D_tcs[92].str)
#define d_CVR  d_tcs[93].str
#define D_CVR (D_tcs[93].str)
#define d_CVN  d_tcs[94].str
#define D_CVN (D_tcs[94].str)
#define d_CTF  d_tcs[95].flg
#define D_CTF (D_tcs[95].flg)
#define d_CXT  d_tcs[96].flg
#define D_CXT (D_tcs[96].flg)
#define d_CG0  d_tcs[97].flg
#define D_CG0 (D_tcs[97].flg)
#define d_CS0  d_tcs[98].str
#define D_CS0 (D_tcs[98].str)
#define d_CE0  d_tcs[99].str
#define D_CE0 (D_tcs[99].str)
#define d_CC0  d_tcs[100].str
#define D_CC0 (D_tcs[100].str)
#define d_AS  d_tcs[101].str
#define D_AS (D_tcs[101].str)
#define d_AE  d_tcs[102].str
#define D_AE (D_tcs[102].str)
#define d_AC  d_tcs[103].str
#define D_AC (D_tcs[103].str)
#define d_EA  d_tcs[104].str
#define D_EA (D_tcs[104].str)
#define d_CXC  d_tcs[105].str
#define D_CXC (D_tcs[105].str)
#define T_CAPS 106
#define d_CKM  d_tcs[106].str
#define D_CKM (D_tcs[106].str)
#define T_FEXTRA 120
#define T_BACKTAB 140
#define T_NAVIGATE 158
#define T_NAVIGATE_DELETE 165
#define T_CURSOR 166
#define T_KEYPAD 170
#define T_OCAPS 188
#define T_ECAPS 201
#define T_N 201

#ifdef MAPKEYS
#  define KMAPDEFSTART 106
#  define NKMAPDEF 82
#  define KMAPADEFSTART 166
#  define NKMAPADEF 22
#  define KMAPMDEFSTART 106
#  define NKMAPMDEF 64
#endif
